from os import environ
import sys
import click
import logging

from flask.cli import FlaskGroup, with_appcontext
from flaschengeist import logger
from flaschengeist.app import create_app

LOGGING_MIN = 5  # TRACE (custom)
LOGGING_MAX = logging.ERROR


def get_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    import platform
    from werkzeug import __version__ as werkzeug_version
    from flask import __version__ as flask_version
    from flaschengeist import __version__

    click.echo(
        f"Python {platform.python_version()}\n"
        f"Flask {flask_version}\n"
        f"Werkzeug {werkzeug_version}\n"
        f"Flaschengeist {__version__}",
        color=ctx.color,
    )
    ctx.exit()


def configure_logger(level):
    """Reconfigure main logger"""
    global logger

    # Handle TRACE -> meaning enable debug even for werkzeug
    if level == 5:
        level = 10
        logging.getLogger("werkzeug").setLevel(level)

    logger.setLevel(level)
    environ["FG_LOGGING"] = logging.getLevelName(level)
    for h in logger.handlers:
        if isinstance(h, logging.StreamHandler) and h.name == "wsgi":
            h.setLevel(level)
            h.setStream(sys.stderr)


@with_appcontext
def verbosity(ctx, param, value):
    """Callback: Toggle verbosity between ERROR <-> TRACE"""

    if not value or ctx.resilient_parsing:
        return
    configure_logger(LOGGING_MAX - max(LOGGING_MIN, min(value * 10, LOGGING_MAX - LOGGING_MIN)))


@click.group(
    cls=FlaskGroup,
    add_version_option=False,
    add_default_commands=False,
    create_app=create_app,
)
@click.option(
    "--version",
    help="Show the flask version",
    expose_value=False,
    callback=get_version,
    is_flag=True,
    is_eager=True,
)
@click.option(
    "--verbose",
    "-v",
    help="Increase logging level",
    callback=verbosity,
    count=True,
    expose_value=False,
)
def cli():
    """Management script for the Flaschengeist application."""
    pass


def main(*args, **kwargs):
    from .plugin_cmd import plugin
    from .export_cmd import export
    from .docs_cmd import docs
    from .run_cmd import run
    from .install_cmd import install

    # Override logging level
    environ.setdefault("FG_LOGGING", logging.getLevelName(LOGGING_MAX))

    cli.add_command(export)
    cli.add_command(docs)
    cli.add_command(install)
    cli.add_command(plugin)
    cli.add_command(run)
    cli(*args, **kwargs)
