import re
import secrets

from io import BytesIO
from typing import Optional
from sqlalchemy import exc
from sqlalchemy_utils import merge_references
from datetime import datetime, timedelta, timezone
from flask.helpers import send_file
from werkzeug.exceptions import NotFound, BadRequest, Forbidden

from .. import logger
from ..config import config
from ..database import db
from ..models import Notification, User, Role
from ..models.user import _PasswordReset
from ..utils.hook import Hook
from ..utils.datetime import from_iso_format
from ..controller import imageController, messageController, pluginController, sessionController
from ..plugins import AuthPlugin


def __active_users():
    """Return query for not deleted users"""
    return User.query.filter(User.deleted == False)


def _generate_password_reset(user):
    """Generate a password reset link for the user"""
    reset = _PasswordReset.query.get(user.id_)
    if not reset:
        reset = _PasswordReset(_user_id=user.id_)
        db.session.add(reset)

    expires = datetime.now(tz=timezone.utc)
    if not reset.expires or reset.expires < expires:
        expires = expires + timedelta(hours=12)
        reset.expires = expires
        reset.token = secrets.token_urlsafe(24)
        db.session.commit()

    return reset


def get_provider(userid: str) -> AuthPlugin:
    return [p for p in pluginController.get_authentication_provider() if p.user_exists(userid)][0]


@Hook
def update_user(user: User, backend: Optional[AuthPlugin] = None):
    """Update user data from backend

    This is seperate function to provide a hook"""
    if not backend:
        backend = get_provider(user.userid)
    backend.update_user(user)
    if not user.display_name:
        user.display_name = "{} {}.".format(user.firstname, user.lastname[0])
    db.session.commit()


def login_user(username, password):
    logger.info("login user {{ {} }}".format(username))
    for provider in pluginController.get_authentication_provider():
        uid = provider.login(username, password)
        if isinstance(uid, str):
            user = get_user(uid)
            if not user:
                logger.debug("User not found in Database.")
                user = User(userid=uid)
                db.session.add(user)
            update_user(user, provider)
            return user
    return None


def request_reset(user: User):
    logger.debug(f"New password reset request for {user.userid}")
    reset = _generate_password_reset(user)

    subject = str(config["MESSAGES"]["password_subject"]).format(name=user.display_name, username=user.userid)
    text = str(config["MESSAGES"]["password_text"]).format(
        name=user.display_name,
        username=user.userid,
        link=f'https://{config["FLASCHENGEIST"]["domain"]}/reset?token={reset.token}',
    )
    messageController.send_message(messageController.Message(user, text, subject))


def reset_password(token: str, password: str):
    if len(token) != 32:
        raise BadRequest

    reset = _PasswordReset.query.filter(_PasswordReset.token == token).one_or_none()
    logger.debug(f"Token is {'valid' if reset else 'invalid'}")
    if not reset or reset.expires < datetime.now(tz=timezone.utc):
        raise Forbidden

    modify_user(reset.user, None, password)
    sessionController.delete_sessions(reset.user)

    db.session.delete(reset)
    db.session.commit()


def set_roles(user: User, roles: list[str], create=False):
    """Set roles of user

    Args:
        user: User to set roles of
        roles: List of role names
        create: If set to true, create not existing roles
    Raises:
        BadRequest if invalid arguments given or not all roles found while *create* is set to false
    """
    from .roleController import create_role

    if not isinstance(roles, list) and any([not isinstance(r, str) for r in roles]):
        raise BadRequest("Invalid role name")
    fetched = Role.query.filter(Role.name.in_(roles)).all()
    if len(fetched) < len(roles):
        if not create:
            raise BadRequest("Invalid role name, role not found")
        # Create all new roles
        fetched += [create_role(role_name) for role_name in roles if not any([role_name == r.name for r in fetched])]
    user.roles_ = fetched


def modify_user(user: User, password: str, new_password: str = None):
    """Modify given user on the backend

    Args:
        user: User object to sync with backend
        password: Current password (most backends are needing this)
        new_password (optional): New password, if password should be changed

    Raises:
        NotImplemented: If backend is not capable of this operation
        BadRequest: Password is wrong or other logic issues
    """
    provider = get_provider(user.userid)
    provider.modify_user(user, password, new_password)

    if new_password:
        logger.debug(f"Password changed for user {user.userid}")
        subject = str(config["MESSAGES"]["password_changed_subject"]).format(
            name=user.display_name, username=user.userid
        )
        text = str(config["MESSAGES"]["password_changed_text"]).format(
            name=user.display_name,
            username=user.userid,
        )
        messageController.send_message(messageController.Message(user, text, subject))


def get_users(deleted=False):
    query = __active_users() if not deleted else User.query
    return query.all()


def get_user_by_role(role: Role):
    return User.query.join(User.roles_).filter_by(role_id=role.id).all()


def get_user(uid, deleted=False) -> User:
    """Get an user by userid from database
    Args:
        uid: Userid to search for
        deleted: Set to true to also search deleted users
    Returns:
        User fround
    Raises:
        NotFound if not found"""
    user = (__active_users() if not deleted else User.query).filter(User.userid == uid).one_or_none()
    if not user:
        raise NotFound
    return user


@Hook
def delete_user(user: User):
    """Delete given user"""
    # First let the backend delete the user, as this might fail
    provider = get_provider(user.userid)
    provider.delete_user(user)

    # Clear all easy relationships
    user.avatar_ = None
    user._attributes.clear()
    user.roles_.clear()
    user.sessions_.clear()
    user.reset_requests_.clear()
    # Now move all other references to the DELETED_USER
    try:
        deleted_user = get_user("__deleted_user__", True)
    except NotFound:
        deleted_user = User(
            userid="__deleted_user__", firstname="USER", lastname="DELETED", display_name="DELETED USER", deleted=True
        )
        db.session.add(user)
        db.session.flush()
    merge_references(user, deleted_user)
    db.session.commit()
    # Now try to delete the user for real
    try:
        db.session.delete(user)
        db.session.commit()
    except exc.IntegrityError:
        logger.error("Delete of user failed, there might be ForeignKey contraits from disabled plugins", exec_info=True)
        # Remove at least all personal data
        user.userid = f"__deleted_user__{user.id_}"
        user.display_name = "DELETED USER"
        user.firstname = ""
        user.lastname = ""
        user.deleted = True
        user.birthday = None
        user.mail = None
        db.session.commit()


def register(data, passwd=None):
    """Register a new user
    Args:
        data: dictionary containing valid user properties
        passwd: optional a password, default: 16byte random
    """
    allowed_keys = User().serialize().keys()
    values = {key: value for key, value in data.items() if key in allowed_keys}
    roles = values.pop("roles", [])
    if "birthday" in data:
        values["birthday"] = from_iso_format(data["birthday"]).date()
    if "mail" in data and not re.match(r"[^@]+@[^@]+\.[^@]+", data["mail"]):
        raise BadRequest("Invalid mail given")
    user = User(**values)
    set_roles(user, roles)

    password = passwd if passwd else secrets.token_urlsafe(16)
    try:
        provider = [p for p in pluginController.get_authentication_provider() if p.can_register()][0]
        provider.create_user(user, password)
        db.session.add(user)
        db.session.commit()
    except IndexError as e:
        logger.error("No authentication backend, allowing registering new users, found.")
        raise e
    except exc.IntegrityError:
        raise BadRequest("userid already in use")

    if user.mail:
        reset = _generate_password_reset(user)

        subject = str(config["MESSAGES"]["welcome_subject"]).format(name=user.display_name, username=user.userid)
        text = str(config["MESSAGES"]["welcome_text"]).format(
            name=user.display_name,
            username=user.userid,
            password_link=f'https://{config["FLASCHENGEIST"]["domain"]}/reset?token={reset.token}',
        )
        messageController.send_message(messageController.Message(user, text, subject))

    provider.update_user(user)

    return user


def load_avatar(user: User):
    if user.avatar_ is not None:
        return imageController.send_image(image=user.avatar_)
    else:
        provider = get_provider(user.userid)
        avatar = provider.get_avatar(user)
        if len(avatar.binary) > 0:
            return send_file(BytesIO(avatar.binary), avatar.mimetype)
    raise NotFound


def save_avatar(user, file):
    get_provider(user.userid).set_avatar(user, file)
    db.session.commit()


def delete_avatar(user):
    get_provider(user.userid).delete_avatar(user)
    db.session.commit()


def persist(user=None):
    if user:
        db.session.add(user)
    db.session.commit()


def get_notifications(user, start=None):
    query = Notification.query.filter(Notification.user_id_ == user.id_)
    if start is not None:
        query = query.filter(Notification.time > start)
    return query.order_by(Notification.time).all()


def delete_notification(nid, user):
    Notification.query.filter(Notification.id == nid).filter(Notification.user_ == user).delete()
    db.session.commit()
