from __future__ import annotations  # TODO: Remove if python requirement is >= 3.12 (? PEP 563 is defered)

from typing import Any, List, Dict
from sqlalchemy.orm.collections import attribute_mapped_collection

from ..database import db
from ..database.types import Serial


class PluginSetting(db.Model):
    __allow_unmapped__ = True
    __tablename__ = "plugin_setting"
    id = db.Column("id", Serial, primary_key=True)
    plugin_id: int = db.Column("plugin", Serial, db.ForeignKey("plugin.id"))
    name: str = db.Column(db.String(127), nullable=False)
    value: Any = db.Column(db.PickleType(protocol=4))


class BasePlugin(db.Model):
    __allow_unmapped__ = True
    __tablename__ = "plugin"
    id: int = db.Column("id", Serial, primary_key=True)
    name: str = db.Column(db.String(127), nullable=False)
    """Name of the plugin, loaded from distribution"""
    installed_version: str = db.Column("version", db.String(30), nullable=False)
    """The latest installed version"""
    enabled: bool = db.Column(db.Boolean, default=False)
    """Enabled state of the plugin"""
    permissions: List["Permission"] = db.relationship(
        "Permission", cascade="all, delete, delete-orphan", back_populates="plugin_", lazy="select"
    )
    """Optional list of custom permissions used by the plugin
    
    A good style is to name the permissions with a prefix related to the plugin name,
    to prevent clashes with other plugins. E. g. instead of *delete* use *plugin_delete*.
    """

    __settings: Dict[str, "PluginSetting"] = db.relationship(
        "PluginSetting",
        collection_class=attribute_mapped_collection("name"),
        cascade="all, delete, delete-orphan",
        lazy="subquery",
    )

    def get_setting(self, name: str, **kwargs):
        """Get plugin setting

        Args:
            name: string identifying the setting
            default: Default value
        Returns:
            Value stored in database (native python)
        Raises:
            `KeyError` if no such setting exists in the database
        """
        try:
            return self.__settings[name].value
        except KeyError as e:
            if "default" in kwargs:
                return kwargs["default"]
            raise e

    def set_setting(self, name: str, value):
        """Save setting in database

        Args:
            name: String identifying the setting
            value: Value to be stored
        """
        if value is None and name in self.__settings.keys():
            del self.__settings[name]
        else:
            setting = self.__settings.setdefault(name, PluginSetting(plugin_id=self.id, name=name, value=None))
            setting.value = value
